/***********************************************************************************************************************
 *
 * SolidBlue - open source safe data
 * Copyright (C) 2011-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://solidblue.java.net
 * SCM: https://bitbucket.org/tidalwave/solidblue-src
 *
 **********************************************************************************************************************/
package it.tidalwave.actor.impl;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.io.Serializable;
import com.eaio.uuid.UUID;
import org.joda.time.DateTime;
import org.joda.time.Duration;
import it.tidalwave.actor.Collaboration;
import it.tidalwave.actor.CollaborationCompletedMessage;
import it.tidalwave.actor.CollaborationStartedMessage;
import it.tidalwave.actor.annotation.Message;
import lombok.Getter;
import lombok.EqualsAndHashCode;
import lombok.RequiredArgsConstructor;
import lombok.ToString;
import lombok.extern.slf4j.Slf4j;

/***********************************************************************************************************************
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@RequiredArgsConstructor @EqualsAndHashCode(of="id") @Slf4j @ToString(of={"id"})
public class DefaultCollaboration implements Serializable, Collaboration
  {
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    private final static DefaultCollaboration NULL = new DefaultCollaboration(new Object())
      {
        @Override
        public void bindToThread() 
          {
          }

        @Override
        public void unbindFromThread() 
          {
          }
      };
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    @RequiredArgsConstructor @ToString
    private static class IdentityWrapper
      {
        @Nonnull
        private final Object object;

        @Override
        public boolean equals (Object object) 
          {
            if ((object == null) || (getClass() != object.getClass()))
              {
                return false;
              }
            
            final IdentityWrapper other = (IdentityWrapper)object;
            return this.object == other.object;
          }

        @Override
        public int hashCode() 
          {
            return object.hashCode();
          }
      }
    
    private final static ThreadLocal<DefaultCollaboration> THREAD_LOCAL = new ThreadLocal<DefaultCollaboration>();
    
    private final UUID id = new UUID();
    
    @Nonnull @Getter
    private final Object originatingMessage;
    
    private final long startTime = System.currentTimeMillis();
    
    private boolean completed;
    
    /** List of threads currently working for this Collaboration. */
    // No need for being weak, since objects are explicitly removed
    private final List<Thread> runningThreads = new ArrayList<Thread>();
    
    /** List of messages currently being delivered as part of this Collaboration. */
    // No need for being weak, since objects are explicitly removed
    private final List<Object> deliveringMessages = new ArrayList<Object>();
    
    /** List of messages pending to be consumed as part of this Collaboration. */
    // No need for being weak, since objects are explicitly removed
    private final List<IdentityWrapper> pendingMessages = new ArrayList<IdentityWrapper>();
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override @Nonnull
    public DateTime getStartTime()
      {
        return new DateTime(startTime);
      }
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override @Nonnull
    public Duration getDuration()
      {
        return new Duration(startTime, System.currentTimeMillis());
      }
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override
    public synchronized void waitForCompletion() 
      throws InterruptedException
      {
        while (!isCompleted())
          {
            wait();  
          }
      }
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override
    public synchronized boolean isCompleted()
      {
        return completed;
      }
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override @Nonnegative
    public synchronized int getDeliveringMessagesCount()
      {
        return deliveringMessages.size();  
      }
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override @Nonnegative
    public synchronized int getPendingMessagesCount()
      {
        return pendingMessages.size();  
      }
    
    /*******************************************************************************************************************
     * 
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override @Nonnegative
    public synchronized int getRunningThreadsCount()
      {
        return runningThreads.size();  
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    @Nonnull
    public static DefaultCollaboration getCollaboration (final @Nonnull Object object)
      {
        return (object instanceof Provider) ? (DefaultCollaboration)((Provider)object).getCollaboration() : NULL;
      }

    /*******************************************************************************************************************
     * 
     * Gets the {@link Collaboration} bound to the current thread or creates a new one.
     * 
     * @return  the {@code Collaboration}
     * 
     ******************************************************************************************************************/
    @Nonnull
    public static DefaultCollaboration getOrCreateCollaboration (final @Nonnull Object source)
      {
        DefaultCollaboration collaboration = THREAD_LOCAL.get();
        
        if (collaboration == null) 
          {
            collaboration = new DefaultCollaboration(source);
            CollaborationStartedMessage.forCollaboration(collaboration).send();
          }
        
        return collaboration;
      } 

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public synchronized void bindToThread() 
      {
        log.trace("bindToThread()");
        THREAD_LOCAL.set(this);
        runningThreads.add(Thread.currentThread());
        notifyAll();
        log();
      }

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public synchronized void unbindFromThread() 
      {
        log.trace("unbindFromThread()");
        runningThreads.remove(Thread.currentThread());
        THREAD_LOCAL.remove();
        notifyAll();
        log();
        eventuallySendCompletionMessage(null);    
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public synchronized void registerDeliveringMessage (final @Nonnull Object message)
      {
        log.trace("registerDeliveringMessage({})", message);
        
        if (message.getClass().getAnnotation(Message.class).daemon())
          {
            deliveringMessages.add(message);  
            notifyAll();
            log();
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public synchronized void unregisterDeliveringMessage (final @Nonnull Object message)
      {
        log.trace("unregisterDeliveringMessage({})", message);
        
        if (message.getClass().getAnnotation(Message.class).daemon())
          {
            deliveringMessages.remove(message);  
            notifyAll();
            log();
            eventuallySendCompletionMessage(message);    
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public synchronized void registerPendingMessage (final @Nonnull Object message)
      {
        log.trace("registerPendingMessage({})", message);
        
        if (message.getClass().getAnnotation(Message.class).daemon())
          {
            pendingMessages.add(new IdentityWrapper(message));  
            notifyAll();
            log();
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public synchronized void unregisterPendingMessage (final @Nonnull Object message)
      {
        log.trace("unregisterPendingMessage({})", message);
        
        if (message.getClass().getAnnotation(Message.class).daemon())
          {
            pendingMessages.remove(new IdentityWrapper(message));  
            notifyAll();
            log();
            eventuallySendCompletionMessage(message);    
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    private void eventuallySendCompletionMessage (final @Nonnull Object message)
      {
        if (!completed && (deliveringMessages.size() + pendingMessages.size() + runningThreads.size() == 0))
          {
            log.debug(">>>> sending completion message for {}", this);
            completed = true;
            THREAD_LOCAL.remove();
            CollaborationCompletedMessage.forCollaboration(this).send();  
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    private void log() // FIXME: drop or move out of synchronized
      {
//        log.trace("{}: delivering messages: {}, pending messages: {}, running threads: {}",
//                  new Object[] {this, deliveringMessages.size(), pendingMessages.size(), runningThreads.size()});  
//        
//        if (pendingMessages.size() < 2)
//          {
//            log.info(">>>> pending messages: {}", pendingMessages);  
//          }  
      }
  }
