/***********************************************************************************************************************
 *
 * SolidBlue - open source safe data
 * Copyright (C) 2011-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://solidblue.java.net
 * SCM: https://bitbucket.org/tidalwave/solidblue-src
 *
 **********************************************************************************************************************/
package it.tidalwave.actor.netbeans.impl;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.ThreadSafe;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.actor.impl.DefaultCollaboration;
import it.tidalwave.eventbus.EventBus;
import it.tidalwave.eventbus.EventBusListener;
import lombok.extern.slf4j.Slf4j;

/***********************************************************************************************************************
 *
 * An implementation of {@link EventBus}.
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@ServiceProvider(service=EventBus.class)
@ThreadSafe @Slf4j
public class CollaborationAwareEventBus implements EventBus
  {
    private final Map<Class<?>, List<WeakReference<EventBusListener<?>>>> listenersMapByTopic =
            new HashMap<Class<?>, List<WeakReference<EventBusListener<?>>>>();
    
    private final List<Object> keeperAlive = new ArrayList<Object>();
    
//    private final RequestProcessor requestProcessor = RequestProcessor.getDefault();
    private final Executor requestProcessor = Executors.newFixedThreadPool(10);
            
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public <Topic> void publish (final @Nonnull Topic event) 
      {
        publish((Class<Topic>)event.getClass(), event);
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public <Topic> void publish (final @Nonnull Class<Topic> topic, final @Nonnull Topic event) 
      {
        log.trace("publish({}, {})", topic, event);
        requestProcessor.execute(new Runnable() 
          {
            @Override
            public void run() 
              {
                publishImpl(topic, event);
                final DefaultCollaboration collaboration = DefaultCollaboration.getCollaboration(event);
                collaboration.unregisterDeliveringMessage(event);
              }
        });
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public <Topic> void subscribe (final @Nonnull Class<Topic> topic, final @Nonnull EventBusListener<Topic> listener) 
      {
        log.debug("subscribe({}, {})", topic, listener);
        keeperAlive.add(listener); // FIXME 
       findListenersByTopic(topic).add(new WeakReference<EventBusListener<Topic>>(listener));
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void unsubscribe (final @Nonnull EventBusListener<?> listener)
      {
        log.info("unsubscribe({})", listener);

        for (final List<WeakReference<EventBusListener<?>>> list : listenersMapByTopic.values())
          {
            for (final Iterator<WeakReference<EventBusListener<?>>> i = list.iterator(); i.hasNext(); )
              {
                final WeakReference<?> ref = i.next();

                if ((ref.get() == null) || (ref.get() == listener))
                  {
                    i.remove();
                  }
              }
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private <Topic> void publishImpl (final @Nonnull Class<Topic> topic, final @Nonnull Topic event) 
      {
        final List<WeakReference<EventBusListener<Topic>>> listeners = new ArrayList<WeakReference<EventBusListener<Topic>>>(findListenersByTopic(topic));
        
        for (final WeakReference<EventBusListener<Topic>> listenerReference : listeners)
          {
            final EventBusListener<Topic> listener = listenerReference.get();

            if (listener != null)
              {
                try
                  {
                    listener.notify(event);   
                  }
                catch (Throwable t)
                  {
                    log.warn("publish()", t); 
                  }
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private <Topic> List<WeakReference<EventBusListener<Topic>>> findListenersByTopic (final @Nonnull Class<Topic> topic)
      {
        final List tmp = listenersMapByTopic.get(topic);
        List<WeakReference<EventBusListener<Topic>>> listeners = tmp;
        
        if (listeners == null)
          {
            listeners = new ArrayList<WeakReference<EventBusListener<Topic>>>();
            listenersMapByTopic.put(topic, (List)listeners);
          } 
        
        return listeners;
      }
  }
